import type { BaseRoot, LimitLiteral } from "@arktype/schema";
import { type requireKeys } from "@arktype/util";
import type { ParseContext } from "../../../scope.js";
import type { InfixOperator } from "../../semantic/infer.js";
import { Scanner } from "../shift/scanner.js";
import { type Comparator, type MinComparator, type OpenLeftBound, type StringifiablePrefixOperator } from "./shared.js";
type BranchState = {
    prefixes: StringifiablePrefixOperator[];
    leftBound: OpenLeftBound | null;
    intersection: BaseRoot | null;
    union: BaseRoot | null;
};
export type DynamicStateWithRoot = requireKeys<DynamicState, "root">;
export declare class DynamicState {
    readonly ctx: ParseContext;
    readonly scanner: Scanner;
    root: BaseRoot<any> | undefined;
    branches: BranchState;
    finalizer: Scanner.FinalizingLookahead | undefined;
    groups: BranchState[];
    constructor(def: string, ctx: ParseContext);
    error(message: string): never;
    hasRoot(): this is DynamicStateWithRoot;
    setRoot(root: BaseRoot): void;
    unsetRoot(): this["root"];
    constrainRoot(...args: Parameters<BaseRoot<any>["constrain"]>): void;
    finalize(finalizer: Scanner.FinalizingLookahead): void;
    reduceLeftBound(limit: LimitLiteral, comparator: Comparator): void;
    finalizeBranches(): void;
    finalizeGroup(): void;
    addPrefix(prefix: StringifiablePrefixOperator): void;
    applyPrefixes(): void;
    pushRootToBranch(token: "|" | "&"): void;
    parseUntilFinalizer(): DynamicStateWithRoot;
    parseOperator(this: DynamicStateWithRoot): void;
    parseOperand(): void;
    private assertRangeUnset;
    reduceGroupOpen(): void;
    previousOperator(): MinComparator | StringifiablePrefixOperator | InfixOperator | undefined;
    shiftedByOne(): this;
}
export {};
